#
# Copyright 2018 3liz
# Author: David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import tornado
import logging

from .basehandler import BaseHandler
from ..version import __version__
from ..config import config_to_dict

LOGGER=logging.getLogger('SRVLOG')


class PingHandler(BaseHandler):
    
    def set_default_headers(self) -> None:
        super().set_default_headers()
        # Disable cache because this is healtcheck requests
        self.set_header("Cache-control","no-store")
        self.set_header("X-Server-Status","ok")
        
    def get(self):
        self.write_json({ 'status': 'ok' })

    def head(self):
        pass


class StatusHandler(BaseHandler):

    def get(self):

        path = self.request.path.rstrip('/')

        if path == '/status/config':
            response = config_to_dict()
        elif path == '/status/env':
            response = dict(os.environ)
        elif path == '/status/stats':
            response = self.application.stats
        elif path == '/status/versions':
            response = self.get_versions()
        else:
            response = self.get_versions()
            req = self.request
            def _link(path: str, title: str, rel: str):
                return {
                    'href' : f"{req.protocol}://{req.host}{path}",
                    'rel'  : rel,
                    'title': title,
                    'type' : "application/json",
                }
            response.update(links=[
                _link("/status/config" , "Server configuration", "status"),
                _link("/status/env", "Execution environment", "status"),
                _link("/status/versions"  , "Versions", "status"),
                _link("/status/"   , "Server status", "self"),
            ])
                        
        self.write_json(response)

    def get_versions(self):
        try:
            from qgis.core import Qgis
            QGIS_VERSION=Qgis.QGIS_VERSION_INT
            QGIS_RELEASE=Qgis.QGIS_RELEASE_NAME
        except Exception as e:
            LOGGER.error("Cannot get Qgis version %s", e)
            QGIS_VERSION="n/a"
            QGIS_RELEASE="n/a"

        return dict(tornado_ver=tornado.version,
                    version = __version__,
                    qgis_version=QGIS_VERSION,
                    qgis_release=QGIS_RELEASE)



""" AMQP monitor for qgis requests
"""
import logging

from .config  import confservice

LOGGER = logging.getLogger('SRVLOG')

class Monitor:

    @classmethod
    def instance(cls) -> 'Monitor':

        if hasattr(cls,'_instance'):
            return cls._instance

        name = confservice.get('server','monitor', fallback=None)
        service = None
        if name:
            from pyqgisservercontrib.core import componentmanager as cm
            try:
                service = cm.load_entrypoint('py_qgis_server.monitors',name).initialize()
                setattr(cls,'_instance', service)
                LOGGER.info("Using '%s' monitor service", name)
                return service
            except cm.EntryPointNotFoundError:
                LOGGER.error("Failed to load monitor component: %s", name)




# Copyright 2022 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


""" Cache Observer used for testing 
"""
import logging

from datetime import datetime

LOGGER=logging.getLogger('SRVLOG')

notify_data = {}

def init() -> None:
    pass

def observe(key: str, datetime: datetime, insert: bool) -> None:
    LOGGER.debug("*** TEST CACHE OBSERVER: Received update notification for %s %s [Inserted: %s]", key, datetime, insert)
    notify_data[key] = (key,datetime,insert)




""" Monitor utilities
"""
import os

from typing import Dict, Any

TAG_PREFIX_LEGACY = 'AMQP_GLOBAL_TAG_'
TAG_PREFIX = 'QGSRV_MONITOR_TAG_'

def _get_tags( prefix: str ):
    return ((e.partition(prefix)[2],os.environ[e]) for e in os.environ if e.startswith(prefix))

class MonitorBase:

    def __init__( self ):
        """ Return tags defined in environment
        """
        # Get global tags
        tags =  { t:v for (t,v) in _get_tags(TAG_PREFIX) if t }
        tags.update( (t,v) for (t,v) in _get_tags(TAG_PREFIX_LEGACY) if t )
        self.global_tags = tags

    def emit( self, params: Dict[str,Any], meta: Dict ) -> None:
        raise NotImplementedError("Subclasses must implement this")
 



# Copyright 2022 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


""" Cache Observer that send a BAN request
"""
import asyncio
import logging

from datetime import datetime
from tornado.httpclient import AsyncHTTPClient, HTTPRequest

from pyqgisserver.config import confservice

LOGGER=logging.getLogger('SRVLOG')

server_address = None
http_client    = None

def init() -> None:
    """
    """
    LOGGER.debug("*** Initializing ban observer")
    confservice.add_section('cache.observers:ban')

    global server_address, http_client
    server_address = confservice.get('cache.observers:ban','server_address')
    http_client = AsyncHTTPClient()

    LOGGER.debug("*** Ban observer: sending_request to %s", server_address)

async def ban( key: str) -> None:
    """ Ban key
    """
    LOGGER.info("Sending BAN request to %s", server_address)

    request = HTTPRequest(server_address, method='BAN', 
                          headers={ 'X-Map-Id': key },
                          user_agent="py-qgis-server; ban observer",
                          allow_nonstandard_methods=True)

    response = await http_client.fetch(request, raise_error=False)
    if response.code != 200:
        LOGGER.error("Ban server returned status code %s", response.code)
    


def observe(key: str, datetime: datetime, inserted: bool) -> None:
    asyncio.create_task(ban(key))




#
# Copyright 2018 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" Qgis server request adapters

    Embedded qgis server in a 0MQ worker 

    see: 
        - https://qgis.org/pyqgis/master/server/QgsBufferServerResponse.html
        - https://qgis.org/pyqgis/master/server/QgsBufferServerRequest.html
"""
import os
import logging
import traceback
import hashlib

try:
    import psutil
except ImportError:
    psutil = None

from datetime import datetime
from time import time
from typing import Dict, Optional, Any, Tuple, Callable

from qgis.PyQt.QtCore import Qt, QBuffer, QIODevice, QByteArray
from qgis.core import QgsProject
from qgis.server import (QgsServerRequest,
                         QgsServerResponse,
                         QgsServerException)

from .zeromq.worker import RequestHandler, run_worker
from .qgscache.cachemanager import (get_cacheservice,
                                    preload_projects,
                                    StrictCheckingError,
                                    UnreadableResourceError,
                                    PathNotAllowedError,
                                    UpdateState,
                                    CacheType)

from .qgscache.observer import Client as CacheObserver

from .plugins import load_plugins
from .config import confservice, configure_qgis_api, qgis_api_endpoints

LOGGER = logging.getLogger('SRVLOG')

HTTP_METHODS = {
    'GET' : QgsServerRequest.GetMethod,
    'PUT' : QgsServerRequest.PutMethod,
    'POST': QgsServerRequest.PostMethod,
    'HEAD': QgsServerRequest.HeadMethod,
    'DELETE': QgsServerRequest.DeleteMethod,
    'PATCH': QgsServerRequest.PatchMethod,
}

class Request(QgsServerRequest):

    def __init__(self, handler: RequestHandler ) -> None:
        """ Create a new QgsServerRequest from zmq handler request
        """
        req = handler.request
        
        # Recreate URL
        location = req.headers.get('X-Forwarded-Url',"")
        query    = req.query.lstrip('?')
        if query:
            location += f"?{query}"

        self._data = req.data
        super().__init__(location, HTTP_METHODS[req.method], headers=req.headers)
         
    def data(self) -> QByteArray: 
        """ Return post/put data a QByteArray
        """
        # Make sure that data is valid
        return QByteArray(self._data) if self._data else QByteArray()


class Response(QgsServerResponse):
    """ Adaptor to handler response

        The data is written at 'flush()' call.
    """

    def __init__(self, handler: RequestHandler, metadata_fn: Callable ) -> None:
        super().__init__()
        self._handler = handler
        self._buffer = QBuffer()
        self._buffer.open(QIODevice.ReadWrite)
        self._numbytes = 0
        self._finish   = False
        self._metadata_fn  = metadata_fn

        self._extra_headers = {}

    def get_metadata(self):
        if self._metadata_fn:
            return self._metadata_fn()
        
    def setExtraHeader( self, key: str, value: str ) -> None:
        # Keep extra headers so we may 
        # set them again on clear()
        self._extra_headers[key] = value
        self.setHeader(key, value)

    def setStatusCode(self, code: int) -> None:
        if not self._handler.header_written:
            self._handler.status_code = code
        else:
            LOGGER.error("Cannot set status code after header written")

    def statusCode(self) -> int:
        return self._handler.status_code

    def finish(self) -> None:
        """ Terminate the request
        """
        self._finish = True
        self.flush()

    def flush(self) -> None:
        """ Write the data to the handler buffer 
            and flush the socket

            Headers will be written at the first call to flush()
        """
        try:
            meta = self.get_metadata()

            self._buffer.seek(0)
            bytesAvail = self._buffer.bytesAvailable()
            LOGGER.debug("%s: Flushing response data: (%d bytes)",self._handler.identity, bytesAvail)
            if self._finish and bytesAvail:
                # Make sure that we have Content-length set
                self._handler.headers['Content-Length']=bytesAvail
            # Take care of the logic: if finish and not handler.header_written then there is no
            # chunk following
            send_more = not self._finish or self._handler.header_written
            if bytesAvail:
                LOGGER.debug("Sending bytes %s (send_more: %s)", bytesAvail, send_more)
                self._handler.send( bytes(self._buffer.data()), send_more, meta )
                self._buffer.buffer().clear()
            else:
                # Return empty response
                LOGGER.debug("Sending empty response (send_more: %s)", send_more)
                self._handler.send( b'', send_more, meta )
            # push the sentinel
            if send_more and self._finish:
                LOGGER.debug("Sending EOF")
                self._handler.send( b'', False, meta )
        except Exception:
            trace = traceback.format_exc()
            LOGGER.error("Caught Exception (worker: %s, msg: %s):\n%s",
                         self._handler.identity, self._handler.msgid,
                         trace)
            del self._handler.headers['Content-Type']
            self.sendError(500, "Internal server error")

    def header(self, key: str) -> str:
        return self._handler.headers.get(key)

    def headers(self) -> Dict[str,str]:
        """ Return headers as dict
        """
        return self._handler.headers
        
    def io(self) -> QIODevice:
        return self._buffer

    def data(self) -> QByteArray:
        """ Return buffer data
        """
        return self._buffer.data()

    def setHeader(self, key: str, value: str) -> None:
        if not self._handler.header_written:
            self._handler.headers[key] = value
        else:
            LOGGER.error("Cannot set header after header written")

    def removeHeader(self, key: str) -> None:
        self._handler.headers.pop(key,None)

    def sendError(self, code: int, message: Optional[str]=None) -> None:
        try:
            if not self._handler.header_written:
                LOGGER.error("%s (%s)", message, code)
                self._handler.status_code = code
                self._handler.send(str(message).encode())
                self._finish = True
            else:
                LOGGER.error("Cannot set error after header written")
        except Exception:
            LOGGER.critical("Unrecoverable exception:\n%s", traceback.format_exc())


    def _clearHeaders(self) -> None:
        """ Clear headers set so far
        """
        self._handler.headers = {}
        self._handler.headers.update(self._extra_headers)
 
    def clear(self) -> None:
        self._clearHeaders()
        self.truncate()

    def headersSent(self) -> bool:
        return self._handler.header_written

    def truncate(self) -> None:
        """ Truncate buffer
        """
        self._buffer.seek(0)
        self._buffer.buffer().clear()


class QgsRequestHandler(RequestHandler):

    @classmethod
    def init_server(cls) -> None:
        """ Initialize qgis server
        """
        if hasattr(cls, 'qgis_server' ):
            return
        
        from .utils.qgis import init_qgis_server

        # Enable qgis server verbosity
        if LOGGER.isEnabledFor(logging.DEBUG):
            os.environ['QGIS_SERVER_LOG_LEVEL'] = '0'
            os.environ['QGIS_DEBUG'] = '1'

        cache_config = confservice['projects.cache']
        trust_mode_on = cache_config.getboolean('trust_layer_metadata')
        if trust_mode_on:
            os.environ['QGIS_SERVER_TRUST_LAYER_METADATA'] = 'yes'

        if cache_config.getboolean('disable_getprint'):
            os.environ['QGIS_SERVER_DISABLE_GETPRINT'] = 'yes'

        # Get refresh interval
        cls._cache_service        = get_cacheservice()
        cls._cache_check_interval = cache_config.getint('check_interval')
        cls._cache_last_check     = time()  
      
        cls._pid = os.getpid()
        cls._advanced_report = cache_config.getboolean('advanced_report')

        if cls._advanced_report:
            if psutil is not None:
                cls._process = psutil.Process(cls._pid)
            else:
                LOGGER.warning("No 'PSUtil' module, advanced report will be disabled")
                cls._advanced_report = False
                cls._process = None

        # Attach cache observer
        if cache_config.getboolean('has_observers'):
            LOGGER.info("Attaching worker cache observer")
            cls._cache_observer = CacheObserver()
            cls._cache_service.add_observer(cls._cache_observer.observe)

        # Configure qgis api
        for name,_ in qgis_api_endpoints(enabled_only=False):
            configure_qgis_api(name)                         
        
        verbose = LOGGER.level<=logging.DEBUG or confservice.getboolean('logging','qgis_info')

        LOGGER.debug("Initializing qgis server")
        qgsserver = init_qgis_server( enable_processing=False, 
                                      logger=LOGGER, 
                                      verbose=verbose)

        serverIface = qgsserver.serverInterface()
        load_plugins(serverIface)

        if confservice['management'].getboolean('enabled'):
            from .management.apis import register_management_apis
            register_management_apis(serverIface)

        preload_projects()

        setattr(cls, 'qgis_server' , qgsserver )

    @classmethod
    def refresh_cache(cls) -> None:
        if cls._cache_check_interval <= 0 or time() - cls._cache_last_check < cls._cache_check_interval:
            return
        LOGGER.debug("Refreshing cache")
        iface = cls.qgis_server.serverInterface()
        for (key, state) in cls._cache_service.refresh():
            # Remove from Qgis server ConfigCache
            if state == UpdateState.UPDATED:
                details = cls._cache_service.peek(key)
                iface.removeConfigCacheEntry(details.project.fileName())
        
        cls._cache_last_check = time()

    @classmethod
    def cache_lookup(cls, key) -> Tuple[QgsProject, UpdateState]:
        return cls._cache_service.lookup(key, refresh = cls._cache_check_interval <= 0)

    @classmethod
    def post_process(cls, idle: bool) -> None:
        """ Post processing operation

            At this time request has been replied and worker is not busy anymore
        """
        cls.refresh_cache()

    @classmethod
    def get_modified_time(cls, key: str, from_cache: bool=True) -> datetime:
        return cls._cache_service.get_modified_time(key, from_cache=from_cache)

    def compute_etag(self, uri: str, last_modified: datetime, request: QgsServerRequest) -> Optional[str]:
        """ Compute ETAG for GetCapabilities requests
        """
        conf = confservice['projects.cache']
        if conf.getboolean('force_etag') or conf.getboolean('trust_layer_metadata'):
            owsrequest = request.parameter('REQUEST') or ""
            if owsrequest.lower() == "getcapabilities":
                hasher = hashlib.sha1()
                hasher.update(request.parameter('SERVICE').lower().encode())
                hasher.update((request.parameter('VERSION') or "").lower().encode())
                hasher.update(uri.encode())
                hasher.update(last_modified.isoformat().encode())
                return '"%s"' % hasher.hexdigest()

    def set_etag_header(self, uri: str, last_modified: datetime, request: QgsServerRequest,
                        response: Response ) -> Optional[str]:
        """ Compute and set etag
        """
        computed_etag = self.compute_etag(uri, last_modified, request )
        if computed_etag:
            response.setExtraHeader("Etag", computed_etag)

        return computed_etag

    def check_etag_header(self, uri: str, last_modified: datetime, request: QgsServerRequest,
                          response: Response) -> bool:
        """ Compute etag and check header
        """
        computed_etag = self.set_etag_header(uri, last_modified, request, response)
        if computed_etag is None:
            return False

        # Check etag header
        etag = self.request.headers.get("If-None-Match","")
        return etag == "*" or etag == computed_etag

    @staticmethod
    def run( router: str, identity: str="", **kwargs: Any) -> None:
        """ Run qgis server worker loop
        """
        QgsRequestHandler.init_server()

        run_worker(router, QgsRequestHandler, identity=bytes(identity.encode('ascii')), 
                   postprocess=QgsRequestHandler.post_process, **kwargs)

    QGIS_NO_MAP_ERROR_MSG = "No project defined. For OWS services: please provide a SERVICE and a MAP parameter" 

    def init_metadata_report(self) -> Dict:
        if self._advanced_report:
            start_mem = self._process.memory_info().rss
            def _metadata_report():
                return {
                    'pid'     : self._pid,
                    'mem_used': self._process.memory_info().rss - start_mem,
                }
            return _metadata_report

    def handle_message(self) -> None:
        """ Override this method to handle_messages
        """
        LOGGER.debug("Handling request: %s", self.msgid)

        project_location = self.request.headers.pop('X-Map-Location', None)
        ogc_scheme       = self.request.headers.pop('X-Ogc-Scheme'  , None)

        metadata_report = self.init_metadata_report()

        request  = Request(self)
        response = Response(self, metadata_report)

        if not project_location:
            # Try to get project from environment
            project_location = os.getenv("QGIS_PROJECT_FILE")
   
        if ogc_scheme == 'OWS': 
            if not project_location and request.parameter('SERVICE'):
                # Prevent qgis for returning 500 when MAP is not defined for
                # OWS services
                LOGGER.error("No project defined for %s", request.parameter('SERVICE'))
                # A project is required
                exception = QgsServerException(self.QGIS_NO_MAP_ERROR_MSG, 400)
                response.write(exception)
                response.finish()
                return

            # Handle HEAD method for OWS requests
            # Note that Qgis Server return 501 on head method on OWS methods
            if request.method() == QgsServerRequest.HeadMethod:
                # We can compute etag without loading resource
                if project_location:
                    last_modified = self.get_modified_time(project_location, from_cache=False)
                    self.set_etag_header(project_location, last_modified, request, response)
                    response.finish()
                    return

        self.handle_qgis_request(ogc_scheme, project_location, request, response)

    def handle_qgis_request(self, ogc_scheme: str, project_location: str, request: Request, response: Response ) -> None:
        """ Handle request passed to Qgis
        """
        if not project_location:
            # Pass request directly
            self.qgis_server.handleRequest(request, response)
            return

        # Handle cached project
        iface = self.qgis_server.serverInterface()
        try:
            project, updated = self.cache_lookup(project_location)
            config_path = project.fileName()
            if updated: 
                # Needed to cleanup cached capabilities
                LOGGER.debug("Cleaning config cache entry %s", config_path)
                iface.removeConfigCacheEntry(config_path)

            last_modified = self.get_modified_time(project_location)

            # Set the project uri in separate header, this
            # is useful for invalidating front-end cache
            response.setExtraHeader('X-Map-Id'      , project_location)
            response.setExtraHeader('Last-Modified' , last_modified.astimezone().isoformat())

            # Check etag for OWS requests
            if ogc_scheme == 'OWS':
                if self.check_etag_header(project_location, last_modified, request, response):
                    response.setStatusCode(304)
                    response.finish()
                    return

        except StrictCheckingError:
            response.sendError(422,f"Invalid layers for project '{project_location}' - strict mode on")
        except UnreadableResourceError:
            response.sendError(422,f"Cannot read project resource '{project_location}'")
        except PathNotAllowedError:
            response.sendError(403,"Project path not allowed")
        except FileNotFoundError:
            response.sendError(404,f"Project '{project_location}' not found")
        else:
            # See https://github.com/qgis/QGIS/pull/9773
            iface.setConfigFilePath(config_path)
            self.qgis_server.handleRequest(request, response, project=project)

    @classmethod
    def get_report(cls):
        report = super(QgsRequestHandler,cls).get_report()

        def _to_json(key: str, project: QgsProject, static: bool):
            return dict(
                key=key,
                filename=project.fileName(),
                last_modified=project.lastModified().toString(Qt.ISODate),
                num_layers=project.count(),
                static=static,
            )

        items = { k:(d,False) for k,d in get_cacheservice().items(CacheType.LRU) }
        items.update( (k,(d,True)) for (k,d) in get_cacheservice().items(CacheType.STATIC) )
        
        report.update(
            cache=[_to_json(k,d.project,static) for (k,(d,static)) in items.items()]
        )
        return report



def main():
    """ Run as command line interface
    """
    import os
    import sys
    import argparse
    from .version import __manifest__
    from .config  import (confservice, load_configuration, read_config_file, validate_config_path)
    from .logger import setup_log_handler

    parser = argparse.ArgumentParser(description='Qgis Server Worker')
    parser.add_argument('-d','--debug', action='store_true', default=False, help="debug mode") 
    parser.add_argument('-c','--config', metavar='PATH', nargs='?', dest='config',
                        default=None, help="Configuration file")
    parser.add_argument('--proxy-host' , dest="hostaddr", metavar="host"   , default='localhost' , help="router host")   
    parser.add_argument('--identity'   , default="", help="Set worker identity")
    parser.add_argument('--rootdir'    , default=argparse.SUPPRESS, metavar='PATH', help='Path to qgis projects')
    parser.add_argument('--version'    , action='store_true', default=False, help="Return version number and exit")

    args = parser.parse_args()

    def print_version(verbose=False) -> None:
        """ Display version infos
        """
        from .utils.qgis import print_qgis_version
        program = os.path.basename(sys.argv[0])
        print("{program} {version} (build {buildid},commit {commitid})".format(program=program,**__manifest__))
        print_qgis_version(verbose=verbose)

    if args.version:
        print_version(verbose=args.debug)
        sys.exit(1)
    else:
        print_version()

    load_configuration()

    if args.config:
        with open(args.config, mode='rt') as config_file:
            read_config_file(config_file)

    # Override config
    def set_arg( section:str, name:str ) -> None:
        if name in args:
            confservice.set( section, name, str(getattr(args,name)))

    set_arg('projects.cache', 'rootdir')
    set_arg('zmq', 'hostaddr')

    if args.debug:
        # Force debug mode
        confservice.set('logging', 'level', 'DEBUG')

    print_version()

    validate_config_path('projects.cache','rootdir')

    setup_log_handler(confservice.get('logging','level'))
    print("Log level set to {}\n".format(logging.getLevelName(LOGGER.level)), file=sys.stderr)

    broadcastaddr = confservice.get('zmq','broadcastaddr')
    router = confservice.get('zmq', 'bindaddr')

    QgsRequestHandler.run(router, identity=args.identity,
                          broadcastaddr=broadcastaddr)

    print("Qgis worker terminated", file=sys.stderr)






#
# Copyright 2022 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" Cache observer for triggers

    The observer aggregate update notifications from 
    all workers, it prevents triggering the same update
    multiple times.
"""
import os
import asyncio
import zmq.asyncio
import zmq
import traceback
import logging

from datetime import datetime
from typing import Optional, Awaitable, Any, NamedTuple, Iterable, Tuple

from ..zeromq.utils import _get_ipc
from ..config  import confservice
from .types import UpdateState


LOGGER=logging.getLogger('SRVLOG')


class _CacheUpdate(NamedTuple):
    modified_time: datetime
    status: UpdateState


class Client:

    def __init__(self) -> None:
        """ Cache observer client
        """
        address = _get_ipc('cache_observer')

        ctx = zmq.Context.instance()
        self._sock = ctx.socket(zmq.PUSH)
        self._sock.setsockopt(zmq.IMMEDIATE, 1) # Do no queue if no connection
        self._sock.connect(address)
        self._pid = os.getpid()

    def _send(self, data: Any ) -> None: 
        try:
            self._sock.send_pyobj((self._pid, data), flags=zmq.DONTWAIT)
        except zmq.ZMQError as err:
            if err.errno != zmq.EAGAIN:
                LOGGER.error("%s (%s)", zmq.strerror(err.errno), err.errno)

    def close(self) -> None:
        self._sock.close()

    def observe(self, key: str, modified_time: datetime, state: UpdateState):
        """
        """
        self._send((key,modified_time,state))
        

class Server:

    _declared_observers = []
    _enabled = False

    @classmethod
    def declare_observers(cls):
        """ Check out observers

            must be done before forking workers since
            we want to make aware of the observers situation
        """
        names = (name.strip() for name in confservice.get('projects.cache','observers', fallback="").split(','))
        cls._declared_observers = list(name for name in names if name)

        # XXX: Managment use cache observer for listing cached objects
        if cls._declared_observers or confservice.getboolean('management','enabled'):
            cls._enabled = True

        confservice.set('projects.cache','has_observers', 'yes' if cls._enabled else 'no')

    def __init__(self)-> None:
        """ Run Observer

            :param timeout: timeout delay in seconds
        """
        address = _get_ipc('cache_observer')

        self._observers = []
        self._stopped = True
        self._task = None
        self._last_updates = {}

        if self._declared_observers:
            self._load_observers()

        if self._enabled:
            LOGGER.info("Starting cache observer")
            ctx = zmq.asyncio.Context.instance()
            self._sock = ctx.socket(zmq.PULL)
            self._sock.setsockopt(zmq.RCVTIMEO, 1000)
            self._sock.bind( address )
        else:
            self._sock = None

    def _load_observers(self):
        """ Load registered triggers
        """
        from pyqgisservercontrib.core import componentmanager as cm
        def _load_observers():
            for name in self._declared_observers:
                try:
                    LOGGER.debug("*** Loading cache observer '%s'", name)
                    observer = cm.load_entrypoint('py_qgis_server.cache.observers',name)
                    observer.init()
                    yield observer
                except cm.EntryPointNotFoundError:
                    LOGGER.error("Failed to load cache trigger component: %s", name)

        self._observers = list(_load_observers())

    def run(self) -> None:
        if self._enabled:
            self._task = asyncio.ensure_future(self._run_async())

    async def _run_async(self) -> Awaitable[None]:
        """ Run supervisor
        """
        self._stopped = False

        while not self._stopped:
            try:
                pid, (key, modified_time, state) = await self._sock.recv_pyobj()
                LOGGER.debug("*** CACHE OBSERVER: Received update %s for key %s from pid %s", state, key, pid)
               
                # Check if an entry exists already 
                entry = self._last_updates.get(key)
                if entry:
                    do_notify = entry.modified_time < modified_time
                else:
                    # Do trigger in all cases (UPDATED, INSERTED)
                    do_notify = True

                # Update entry
                self._last_updates[key] = _CacheUpdate(modified_time, state)
                   
                if do_notify:
                    self.notify_observers(key, modified_time, state)

            except zmq.ZMQError as err:
                if err.errno != zmq.EAGAIN:
                    LOGGER.error("%s\n%s", zmq.strerror(err.errno), traceback.format_exc())
            except asyncio.CancelledError:
                raise
            except Exception:
                LOGGER.critical("%s", traceback.format_exc())
                raise

    def stop(self) -> None:
        """ Stop the Observer
        """
        if self._stopped:
            return

        LOGGER.info("Stopping cache observer")
        self._stopped = True
        if self._task and not self._task.cancelled():
            self._task.cancel()

    def notify_observers(self, key: str, modified_time: datetime, state = UpdateState ) -> None:
        """ Run registered observers
        """
        for obs in self._observers:
            try:
                obs.observe(key, modified_time, state == UpdateState.INSERTED)
            except Exception:
                LOGGER.critical("Uncaugh error in observer: %s\n%s", obs, traceback.format_exc())

    def find(self, key:str) -> Optional[_CacheUpdate]:
        """ return entry
        """
        return self._last_updates.get(key)

    def items(self) -> Iterable[Tuple[str,_CacheUpdate]]:
        return self._last_updates.items()


def declare_cache_observers() -> None:
    Server.declare_observers()


def start_cache_observer() -> Server:
    server = Server()
    server.run()
    return server






#
# Copyright 2018 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import sys
import asyncio
import logging
import signal

import tornado.web
import tornado.platform.asyncio

from tornado.web import RedirectHandler

from multiprocessing import Process

from .logger import log_request
from .config import confservice, qgis_api_endpoints

from .handlers import (StatusHandler, 
                       OwsHandler, 
                       OAPIHandler,
                       PingHandler, 
                       LandingPage,
                       NotFoundHandler)

from .zeromq import client, broker

from .qgspool import create_poolserver

from .monitor import Monitor
from .stats import Stats
from .qgscache.observer import declare_cache_observers, start_cache_observer


LOGGER=logging.getLogger('SRVLOG')


def configure_handlers( client: client.AsyncClient ) -> [tornado.web.RequestHandler]:
    """ Configure request handlers
    """
    cfg = confservice['server']

    monitor = Monitor.instance()

    ows_kwargs = dict(
        client       = client,
        monitor      = monitor,
        timeout      = cfg.getint('timeout'),
        allowed_hdrs = tuple(k.upper() for k in cfg.get('allow_headers').split(','))
    )

    end = r"(?:\.html|\.json|/?)"

    handlers = [
        (r"/", LandingPage),
        (r"/ping", PingHandler),
    ]

    def add_handler( path, handler, kwargs ):
        handlers.append( (path, handler, kwargs) )

    # Server status page
    if cfg.getboolean('status_page'):
        handlers.append( ("/status/?.*", StatusHandler) )

    def _ows_args( *args, **kwargs ):
        rv = ows_kwargs.copy()
        rv.update( *args, **kwargs )
        return rv

    add_handler(r"/ows/?", OwsHandler, _ows_args(getfeaturelimit=cfg.getint('getfeaturelimit')))

    wfs3_api_endpoints = [
        rf"{end}",
        rf"/collections(?:/[^/]+(?:/items)?)?{end}",
        rf"/conformance{end}",
        rf"/api{end}",
        r"/static/.*",
    ] 

    # XXX DEPRECATED (to be removed in 1.9)
    handlers.append((rf"/ows/wfs3({end}.*)", RedirectHandler, {'url': "/wfs3{0}" }))

    # New scheme
    kw = _ows_args(service='WFS3')
    for endpoint in wfs3_api_endpoints:
        handlers.append((rf"/wfs3{endpoint}", OAPIHandler, kw) )

    #
    # Add qgis api endpoints
    #
    for name, endpoint in qgis_api_endpoints():
        kw = _ows_args(service=name)
        LOGGER.debug("*** Adding API handler for: %s: %s", name, endpoint)        
        handlers.append( (rf"/{endpoint.strip('/')}/.*", OAPIHandler, kw) )

    return handlers


class Application(tornado.web.Application):

    def __init__(self, router: str) -> None:
        """
        """
        identity = confservice['zmq']['identity']
        identity = "{}-{}".format(identity,os.getpid())

        self._broker_client = client.AsyncClient(router, bytes(identity.encode('ascii')))
        self.stats = Stats()

        self.http_proxy = confservice.getboolean('server','http_proxy')

        super().__init__(configure_handlers(self._broker_client),
                         default_handler_class=NotFoundHandler)

    def log_request(self, handler: tornado.web.RequestHandler ) -> None:
        """ Write HTTP requet to the logs
        """
        log_request(handler)        

    def terminate(self) -> None:
        self._broker_client.terminate()


def setuid( username: str) -> None:
    """ setuid to username uid """
    from pwd import getpwnam, getpwuid
    pw = getpwnam(username)
    os.setgid(pw.pw_gid)
    os.setuid(pw.pw_uid)
    LOGGER.info("Setuid to user {} ({}:{})".format(getpwuid(os.getuid()).pw_name, os.getuid(), os.getgid()))


def create_broker_process( ipcaddr: str ) -> Process:
    """ Create a brker process
    """
    cfg = confservice['zmq']

    LOGGER.info("Starting broker process")
    p = Process(target=broker.run_broker, kwargs=dict(
                inaddr   = ipcaddr,
                outaddr  = cfg['bindaddr'],
                maxqueue = cfg.getint('maxqueue'),
                timeout  = cfg.getint('timeout')))
    p.start()
    return p


def configure_ipc_addresses(workers: int) -> str:
    """ Create ipc socket path
    """
    pid = os.getpid()
    ipc_path = f"/tmp/qgssrv/broker/{pid}/"
    os.makedirs(os.path.dirname(ipc_path), exist_ok=True)
    ipcaddr = f"ipc://{ipc_path}0"
    # Use ipc sockets for managed workers
    confservice.set('zmq','ipcpath', ipc_path)
    if workers > 0:
        confservice.set('zmq','bindaddr'     , f"ipc://{ipc_path}pool0")
        confservice.set('zmq','broadcastaddr', f"ipc://{ipc_path}broadcast0")

    return ipcaddr


def create_ssl_options():
    """ Create an ssl context
    """
    import ssl
    cfg = confservice['server']
    ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ssl_ctx.load_cert_chain(cfg['ssl_cert'],cfg['ssl_key'])
    return ssl_ctx


def initialize_middleware( app ): 
    """ Initialize the middleware
    """
    if confservice.getboolean('server','enable_filters'):
        from .middleware import MiddleWareRouter
        router = MiddleWareRouter(app)
    else:
        router = app

    return router


def run_server( port: int, address: str="", user: str=None, workers: int=0) -> None:
    """ Run the server

        :param port: port number
        :param address: interface address to bind to (default to any)
        :user: User to setuid after opening ports (default no setuid)
    """
    import traceback
    from tornado.netutil import bind_sockets
    from tornado.httpserver import HTTPServer

    kwargs = {}

    ipcaddr = configure_ipc_addresses(workers)

    application = None
    server = None

    if user:
        setuid(user)

    worker_pool = None
    broker_pr = None
    cache_observer = None
    management = None

    # Setup ssl config
    if confservice.getboolean('server','ssl'):
        LOGGER.info("SSL enabled")
        kwargs['ssl_options'] = create_ssl_options()

    if confservice.getboolean('server','http_proxy'):
        LOGGER.info("Proxy configuration enabled")
        kwargs['xheaders'] = True

    # Check for declared cache observers
    declare_cache_observers()

    try:
        broker_pr   = create_broker_process(ipcaddr)
        worker_pool = create_poolserver(workers) if workers>0 else None

        # Since python 3.10 and deprecation of `get_event_loop()` 
        # This is now the preferred way to start tornado application
        # See https://www.tornadoweb.org/en/stable/guide/running.html
        async def _main():
            sockets = bind_sockets(port, address=address)

            LOGGER.info("Running server on port %s:%s", address, port)

            nonlocal application
            application = Application(ipcaddr)

            # Init HTTP server
            nonlocal server
            server = HTTPServer(initialize_middleware(application), **kwargs)
            server.add_sockets(sockets)

            # Activate management
            nonlocal management
            if confservice['management'].getboolean('enabled'):
                from .management.server import start_management_server
                management = start_management_server(worker_pool,ipcaddr)
                management.stats = application.stats

            # Initialize pool supervisor
            if worker_pool:
                worker_pool.start_supervisor()

            # Start cache observer
            nonlocal cache_observer
            cache_observer = start_cache_observer()

            if management:
                management.cache_observer = cache_observer

            event = asyncio.Event()

            loop = asyncio.get_running_loop()
            loop.add_signal_handler(signal.SIGTERM, event.set)
            
            LOGGER.info("Starting processing requests")
            # Wait forever until signal is set 
            # see https://www.tornadoweb.org/en/stable/guide/running.html
            await event.wait()

        # XXX This trigger a deprecation warning in python 3.10
        # but there is no clear alternative with tornado atm
        # See https://github.com/tornadoweb/tornado/issues/3033
        asyncio.run(_main())
    except Exception:
        print("Server exited from exception")
        traceback.print_exc()
        exit_code = 1
    except KeyboardInterrupt:
        print("Keyboard Interrupt", flush=True)
        exit_code = 15
    except SystemExit as exc:
        print("Exiting with code:", exc.code, flush=True)
        exit_code = exc.code
    else:
        exit_code = 0

    # Teardown
    if server is not None:
        server.stop()
    if management is not None:
        management.terminate()
        management = None
    if application is not None:
        application.terminate()
        application = None
        print("PID {}: Server instance stopped".format(os.getpid()), flush=True)
    if cache_observer:
        cache_observer.stop()
    if worker_pool:
        print("Stopping workers", flush=True)
        worker_pool.terminate()
    if broker_pr:
        print("Stopping broker", flush=True)
        broker_pr.terminate()
        broker_pr.join()

    print("Server shutdown", flush=True)
    sys.exit(exit_code)





from ..config import confservice

def _get_ipc( name: str ) -> str:
    ipc_path = confservice.get('zmq','ipcpath', fallback=None)
    if ipc_path:
        return f"ipc://{ipc_path}/{name}"
    else:
        # Get alternate tcp configuration 
        return confservice.get('zmq', f'{name}_addr')




#
# Copyright 2018-2019 3liz
# Author: David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" Qgis server handler
"""
import logging
from time import time
from urllib.parse import urlencode

from ..logger import log_rrequest
from ..zeromq.client import RequestTimeoutError, RequestGatewayError, AsyncClient
from ..monitor import Monitor

from .basehandler import BaseHandler

from typing import Any, Optional, Awaitable, List, Dict

LOGGER = logging.getLogger('SRVLOG')

class AsyncClientHandler(BaseHandler):

    """ Proxy to Qgis 0MQ worker
    """
    def initialize(self, client: AsyncClient, timeout: int, 
                   monitor: Optional[Monitor]=None, 
                   allowed_hdrs: List[str]=[] ) -> None:

        super().initialize()

        self._client       = client
        self._timeout      = timeout
        self._monitor      = monitor
        self._stats        = self.application.stats
        self._allowed_hdrs = allowed_hdrs

        self.ogc_scheme   = None

    def encode_arguments(self) -> str:
        return '?'+urlencode({k:v[0] for k,v in self.request.arguments.items()})

    def set_backend_headers(self, headers: Dict) -> None:
        """ Set headers passed to backend
        """
        project_path = self.get_argument('MAP',default=None)
       
        if project_path:
            headers['X-Map-Location']=project_path 
        if self.ogc_scheme:
            headers['X-Ogc-Scheme'] = self.ogc_scheme

        # Pass etag
        headers['If-None-Match'] = self.request.headers.get("If-None-Match", "")

        def copy_headers(pats):
            headers.update((k,v) for k,v in self.request.headers.items() if \
                           any(map(k.upper().startswith,pats)))

        # Copy custom Qgis/Forwarded headers
        # see https://github.com/qgis/QGIS/pull/41333
        copy_headers(self._allowed_hdrs)

    async def handle_request(self, method: str) -> Awaitable[None]:
        reqtime = time()

        try:
            meta = None
            response = None
            delta = None
            query = self.encode_arguments()

            headers = {}
            proxy_url = self.proxy_url()
            if proxy_url:
                # Send the full path to Qgis 
                headers['X-Forwarded-Url']=f"{proxy_url}{self.request.path.lstrip('/')}"

            self.set_backend_headers(headers)

            data = self.request.body

            if self.get_argument('SERVICE', default=None) and  self.has_body_arguments:
                # Do not let qgis server handle url encoded parameters
                data = None
                if method == 'POST':
                    method = 'GET'

            self._stats.num_requests +=1

            response = await self._client.fetch(query=query, method=method, 
                                                headers=headers, data=data,
                                                timeout=self._timeout)
            status = response.status
            hdrs   = response.headers
            delta  = time() - reqtime

            log_rrequest(proxy_url, status, method, query, delta, hdrs)
          
            # Send response
            for k,v in hdrs.items():
                self.set_header(k,v)

            # Send CORS Header
            self.set_access_control_headers()

            if status == 206:
                # Partial response
                self.set_status(200)
                self.write(response.data)
                await self.flush()
                async for chunk in self._client.fetch_more(response, timeout=self._timeout):
                    self.write(chunk)
                    await self.flush()
                delta = time() - reqtime
            elif status == 509:
                self.send_error(status, reason="Server busy, please retry later")
            else:
                self.set_status(status)
                if response.data:
                    # XXX Tornado do no like 304 
                    # with (potentially empty) chunk
                    self.write(response.data)

            meta = response.metadata

        except RequestTimeoutError:
            status = 504
            delta = time() - reqtime
            # Log the request with status code 499 indicating
            # that the request has not returned
            log_rrequest(proxy_url, 499, method, query, delta, {})
            self.send_error(status, reason="Request timeout error")
        except RequestGatewayError:
            status = 502
            delta = time() - reqtime
            # Log the request 
            log_rrequest(proxy_url, 499, method, query, delta, {})
            self.send_error(status, reason="Backend request error")

        if status >= 500:
            self._stats.num_errors +=1

        # Send monitoring info
        self.emit( status, delta, meta or {})

    def emit(self, status: int, response_time: float, meta: Dict) -> None:
        if not self._monitor:
            return

        if meta:
            LOGGER.debug("### Adv. Metrics => %s", meta)

        params = self.get_monitor_params()
        if params:
            params.update(
                # RESPONSE TIME MUST BE IN MILLISECONDS
                RESPONSE_TIME = int(response_time*1000.0),
                RESPONSE_STATUS  = status,
                RESPONSE_MEMUSED = meta.get('mem_used',0)
            )
            self._monitor.emit( params, meta=self.request.headers )

    async def get(self) -> Awaitable[None]:
        """ Handle Get method
        """
        await self.handle_request('GET')

    async def post(self) -> Awaitable[None]:
        """ Handle Post method
        """
        await self.handle_request('POST')

    async def head(self) -> Awaitable[None]:
        """ Handle HEAD method
        """
        await self.handle_request('HEAD')

    def options(self) -> None:
        """ Implement OPTIONS for validating CORS
        """
        self.set_option_headers('GET, POST, OPTIONS')

    def get_monitor_params( self ) -> Optional[Dict[str,Any]]:
        """ Emit monitoring info
        """
        return None



#
# Copyright 2019 3liz
# Author: David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


""" Define middleware filters

    Declare filters as

    @blockingfilter(uri="/ows/")
    def myfilter1( handller: tornado.web.RequestHandler, *args ) -> None:
        ...

    @asyncfilter(pri=100, uri="/ows/")
    async def myfilter2( handller: tornado.web.RequestHandler, *args) -> None:
        ...


    Filters are applied before sending the request, a filter can return response or 
    raise exception based on authorization rules.

    Registering filters:

    Modules implementing filters must register their filters using setup.py entry_points with the 
    key 'pyqgisserver_filters'

    policy_service.add_policy([myfilter1, myfilter2], pri=1000)
"""

import logging
import re

from tornado.httputil import HTTPServerRequest

from typing import (
    Union,
    Tuple,
    Callable, 
    Awaitable, 
    Optional,
)

LOGGER = logging.getLogger('SRVLOG')


class _FilterBase:

    def __init__(self, match: Optional[Union[str,re.Pattern]]=None, repl: Optional[str]=None):
        if isinstance(match, str):
            match = re.compile(match, re.IGNORECASE)
        self.pattern = match
        self.repl = repl
        self.match_args = []
        self.match_kwargs = {}

    def __str__(self) -> str:
        return f"_FilterBase<{hex(id(self))}>(match={self.pattern}, repl={self.repl})"
    def match(self, path: str) -> Tuple[bool,str]:
        """ Check uri against pattern
        """
        if self.pattern:
            match = self.pattern.match(path)
            if match:
                # match.groups() includes both named and
                # unnamed groups, we want to use either groups
                # or groupdict but not both. 
                if self.pattern.groupindex:
                    self.match_kwargs = match.groupdict()
                else:
                    self.match_args = match.groups() 
                if self.repl:
                    path = self.pattern.sub(self.repl, path, count=1)
                return True, path
        else:
            # Match everything
            return True, path

        return False, None

    def __call__(self, fn: Callable):
        self.fn = fn
        return self

    def apply(self, request: HTTPServerRequest) -> Optional[Awaitable]:
        self.fn(request, *self.match_args, **self.match_kwargs)
        return


class policy_filter(_FilterBase):
    """ Decorator for synchronous request filter
    """
    pass




#
# Copyright 2018-2019 3liz
# Author: David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" Qgis server handler
"""
import logging
from urllib.parse import urlencode

from .asynchandler import AsyncClientHandler

from typing import Any, Union, Dict

LOGGER = logging.getLogger('SRVLOG')


def _decode( b: Union[str,bytes] ) -> str:
    if not isinstance(b,str):
        return b.decode('utf-8')
    return b


class OwsHandler(AsyncClientHandler):

    def initialize(self, *args, getfeaturelimit: int=-1,  **kwargs) -> None:
        super().initialize(*args, **kwargs)
        self.getfeaturelimit = getfeaturelimit
        self.ogc_scheme = 'OWS'

    MONITOR_ARGUMENTS = (
        'MAP',
        'SERVICE',
        'REQUEST',
    )

    def get_monitor_params(self) -> Dict[str,Any]:
        """ Override
        """
        args = self.request.arguments
        params = { k:_decode(args.get(k,["__unknown__"])[0]) for k in self.MONITOR_ARGUMENTS }
        return params

    def prepare(self) -> None:
        super().prepare()
        # Replace query arguments to upper case: (it's ok for OWS)
        self.request.arguments = { k.upper():v for (k,v) in self.request.arguments.items() }

    def fix_getfeature(self, arguments: Dict) -> Dict:
        """ Take care of WFS/GetFeature limit

            Qgis does not set a default limit and unlimited
            request may cause issues
        """
        if self.getfeaturelimit > 0 \
                and arguments.get('SERVICE', b'').upper() == b'WFS' \
                and arguments.get('REQUEST', b'').lower() == b'getfeature':

            if arguments.get('VERSION', b'').startswith(b'2.'):
                key = 'COUNT'
            else:
                key = 'MAXFEATURES'

            limit = self.getfeaturelimit
            try:
                actual_limit = int(arguments.get(key,0))
                if actual_limit > 0:
                    limit =  min(limit, actual_limit)
            except ValueError:
                pass
            arguments[key] = str(limit).encode()

        return arguments

    def encode_arguments(self) -> str:
        arguments = {k:v[0] for k,v in self.request.arguments.items()}
        return '?'+urlencode(self.fix_getfeature(arguments))




#
# Copyright 2020 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" Cache manager for Qgis Projects
"""

import logging
import urllib.parse
import traceback

from urllib.parse import urlparse, urlunparse, urljoin, parse_qs
from typing import Any, Tuple, Optional, Sequence, NamedTuple, Callable, Generator
from collections import OrderedDict
from pathlib import Path
from datetime import datetime
from enum import Enum

from ..utils.lru import lrucache
from ..config import confservice

from .types import UpdateState

from qgis.PyQt.QtCore import Qt 
from qgis.core import (Qgis,
                       QgsApplication,
                       QgsProjectBadLayerHandler, 
                       QgsProject, 
                       QgsMapLayer)

from qgis.server import QgsServerProjectUtils

from pyqgisservercontrib.core import componentmanager

# Import default handlers for auto-registration
from .handlers import * # noqa: F403,F401

LOGGER = logging.getLogger('SRVLOG')


class StrictCheckingError(Exception):
    pass


class PathNotAllowedError(Exception):
    pass


class UnreadableResourceError(Exception):
    """ Indicates that the  ressource exists but is not readable
    """
    pass


class CacheDetails(NamedTuple):
    project: QgsProject 
    timestamp: datetime


CACHE_MANAGER_CONTRACTID = '@3liz.org/cache-manager;1'



def _merge_qs( query1: str, query2: str ) -> str:
    """ Merge query1 with query2 but coerce values
        from query1 
    """
    params_1 = parse_qs(query1)
    params_2 = parse_qs(query2)
    params_2.update(params_1)
    return '&'.join('%s=%s' % (k,v[0]) for k,v in params_2.items())


class CacheType(Enum): 
    LRU = 'lru'
    STATIC = 'static'


class QgisStorageHandler:
    """ Handler for handling Qgis supported storage
        throught the `QgsProjectStorage` api.
    """
    def __init__(self):
        pass

    def get_storage_metadata( self, uri: str ):
        # Check out for storage
        storage = QgsApplication.projectStorageRegistry().projectStorageFromUri(uri)
        if not storage:
            LOGGER.error("No project storage found for %s", uri)
            raise FileNotFoundError(uri)
        res, metadata = storage.readProjectStorageMetadata( uri )
        if not res:
            LOGGER.error("Failed to read storage metadata for %s", uri)
            raise FileNotFoundError(uri)
        return metadata

    def get_modified_time( self, url: urllib.parse.ParseResult) -> datetime:
        """ Return the modified date time of the project referenced by its url
        """
        metadata = self.get_storage_metadata(urlunparse(url))
        return metadata.lastModified.toPyDateTime()

    def get_project( self, url: urllib.parse.ParseResult,
                     project: Optional[QgsProject]=None,
                     timestamp: Optional[datetime]=None) -> Tuple[QgsProject, datetime]:
        """ Create or return a project
        """
        uri = urlunparse(url)

        metadata = self.get_storage_metadata(uri)
        modified_time = metadata.lastModified.toPyDateTime()

        if timestamp is None or timestamp < modified_time:
            cachmngr  = componentmanager.get_service('@3liz.org/cache-manager;1')
            project   = cachmngr.read_project(uri)
            timestamp = modified_time

        return project, timestamp


@componentmanager.register_factory(CACHE_MANAGER_CONTRACTID)
class QgsCacheManager:
    """ Handle Qgis project cache 
    """

    StrictCheckingError=StrictCheckingError
    PathNotAllowedError=PathNotAllowedError
    UnreadableResourceError=UnreadableResourceError

    def __init__(self) -> None:
        """ Initialize cache

            :param size: size of the lru cache
        """
        cnf = confservice['projects.cache']

        size = cnf.getint('size')

        self._create_project = QgsProject
        self._lru_cache            = lrucache(size)
        self._static_cache         = OrderedDict()
        self._strict_check         = cnf.getboolean('strict_check')
        self._trust_layer_metadata = cnf.getboolean('trust_layer_metadata')
        self._disable_getprint     = cnf.getboolean('disable_getprint')
        self._disable_owsurls      = cnf.getboolean('disable_owsurls', fallback=False)
        self._aliases = {}
        self._default_scheme = cnf.get('default_handler')
        self._observers = []   

        allowed_schemes = cnf.get('allow_storage_schemes')
        if allowed_schemes != '*':
            allowed_schemes = [s.strip() for s in allowed_schemes.split(',')]
        self._allowed_schemes = allowed_schemes
    
        # Set the base url for file protocol
        self._aliases['file'] = 'file:///%s/' % cnf.get('rootdir').strip('/')

        if self._disable_getprint:
            LOGGER.info("** Cache: Getprint disabled")

        if self._trust_layer_metadata:
            LOGGER.info("** Cache: Trust Layer Metadata on")

        # Load protocol handlers
        componentmanager.register_entrypoints('qgssrv_contrib_protocol_handler')

    def add_observer( self, observer: Callable[[str,datetime,int],None] ) -> None: 
        """ Add observer for cache invalidation
        """
        self._observers.append(observer)

    def notify_observers( self, key: str, modified_time: datetime,
                          state: UpdateState) -> None:
        """ Notify all observers
        """
        if not self._observers:
            return

        modified_time = modified_time.replace(microsecond=0)
        for obs in self._observers:
            try:
                obs(key, modified_time, int(state))
            except Exception:
                LOGGER.critical("Uncaugh error in observer: %s\n%s", obs, traceback.format_exc())

    @property
    def trust_mode_on(self) -> bool:
        return self._trust_layer_metadata

    @property
    def strict_mode_on(self) -> bool:
        return self._strict_check

    def items(self, cachetype: CacheType = CacheType.LRU) -> Sequence[Tuple[str,CacheDetails]]:
        if cachetype == CacheType.LRU:
            return self._lru_cache.items()
        elif cachetype == CacheType.STATIC:
            return self._static_cache.items()

    def resolve_alias(self, key: str ) -> urllib.parse.ParseResult:
        """ Resolve scheme from configuration variables
        """
        url    = urlparse(key)
        scheme = url.scheme or self._default_scheme
        LOGGER.debug("Resolving '%s' protocol for '%s'", scheme, key)
        baseurl = self._aliases.get(scheme)
        if not baseurl:
            try:
                # Check for user-defined scheme
                baseurl = confservice.get('projects.schemes',scheme.replace('-','_').lower())
            except KeyError:
                pass
            else:
                LOGGER.info("Scheme '%s' aliased to %s", scheme, baseurl)
                self._aliases[scheme] = baseurl
        if baseurl:
            if '{path}' in baseurl:
                url = urlparse(baseurl.format(path=url.path))
            else:
                baseurl = urlparse(baseurl)
                # Build a new query from coercing with base url params 
                query = _merge_qs(baseurl.query, url.query)
                # XXX Note that the path of the base url must be terminated by '/'
                # otherwise urljoin() will replace the base name - may be not what we want
                url = urlparse(urljoin(baseurl.geturl(),url.path+'?'+query))
                # Make sure that the result url path is relative to base url
                try:
                    # Ensure that if an absolute path is given, we may extract
                    # a relative path to the base url - note that the base url may
                    # not have a leading '/'
                    Path(url.path).relative_to(Path('/') / baseurl.path)
                except ValueError:
                    LOGGER.error("The path '%s' is outside base path '%s'", url.path, baseurl.path)
                    raise PathNotAllowedError()                    

        return url

    def get_protocol_handler(self, key: str, scheme: Optional[str] ) -> Any:
        """ Find protocol handler for the given scheme
        """
        scheme = scheme or self._default_scheme
        # Check for allowed schemes
        if self._allowed_schemes != '*' and scheme not in self._allowed_schemes:
            LOGGER.error("Scheme %s not allowed", scheme)
            raise PathNotAllowedError(key)
        # Retrieve the protocol-handler
        try:
            store = componentmanager.get_service('@3liz.org/cache/protocol-handler;1?scheme=%s' % scheme)
        except componentmanager.FactoryNotFoundError:
            # Fallback to Qgis storage handler
            store = QgisStorageHandler()

        return store 

    def refresh(self) -> Generator[Tuple[str,UpdateState], None, None]:
        """ Refresh all entries

            keys are returned from most recently user to the last recently,
            then we have to update in reverse for preserving order 
        """
        # Update static cache;
        for key,_ in self.items(CacheType.STATIC):
            self.update_static_entry(key)

        # Update both caches
        keys = reversed([k for k,_ in self.items(CacheType.LRU)])
        return ((key, self.update_entry(key)) for key in keys)

    def peek(self, key:str, cachetype: Optional[CacheType] = None) -> CacheDetails:
        """ Return cache details 
        """
        if cachetype == CacheType.LRU:
            return self._lru_cache.peek(key)
        elif cachetype == CacheType.STATIC:
            return self._static_cache.get(key)
        else:
            return self._lru_cache.peek(key) or self._static_cache.get(key)

    def get_modified_time(self, key: str, from_cache: bool=True) -> datetime:
        """ Get the modified time for the given project uri
        """
        details = self.peek(key)
        if details:
            if from_cache:
                # Return from cached resource
                return details.timestamp.replace(microsecond=0)

            # Trust Qgis to return modified time
            last_modified = details.project.lastModified()
            if not last_modified.isValid():
                # Occurs if resource is not valid 
                LOGGER.error("QgsProject::lastModified() returned invalid date time for %s", 
                             details.project.fileName())
                raise FileNotFoundError(key)
            last_modified = last_modified.toPyDateTime()
        else:
            # Get modified 
            url = self.resolve_alias(key)
            store = self.get_protocol_handler(key, url.scheme)
            last_modified = store.get_modified_time(url)

        return last_modified.replace(microsecond=0)

    def _get_project_details(self, key: str, 
                             details: Optional[CacheDetails]) -> Tuple[CacheDetails, bool]: 
        """ Return updated project details
        """
        url   = self.resolve_alias(key)
        store = self.get_protocol_handler(key, url.scheme)

        if details is not None:
            project, timestamp  = store.get_project(url, **details._asdict())
            update = UpdateState.UPDATED if timestamp != details.timestamp else UpdateState.UNCHANGED
        else:
            project, timestamp = store.get_project(url)
            update = UpdateState.INSERTED

        return CacheDetails(project, timestamp), update

    def update_static_entry( self, key: str ) -> UpdateState:
        """ Update static cache
        """
        details, update = self._get_project_details(key, self._static_cache.get(key))
        self._static_cache[key] = details
        return update

    def update_entry(self, key: str) -> UpdateState:
        """ Update LRU entry
        """
        # Get details for the project
        details = self._lru_cache.peek(key)

        # Lookup in static cache
        if not details and key in self._static_cache:
            details = self._static_cache[key]
            details, update = self._get_project_details(key, details)
            if update == UpdateState.UPDATED:
                # Update the static cache
                self._static_cache[key] = details
            # LRU is updated with new content
            update = UpdateState.INSERTED
        else:
            details, update = self._get_project_details(key, details)

        self._lru_cache[key] = details

        # Notify
        if update:
            LOGGER.info("Updated project '%s' in cache", key),
            self.notify_observers(key, details.timestamp, update)

        return update

    def lookup(self, key: str, refresh: bool = True) -> Tuple[QgsProject, UpdateState]:
        """ Lookup entry from key

            If refresh is False, return actual cache 
            content without refreshing/updating the entry.
        """
        if not refresh:
            # Lookup LRU
            details = self.peek(key, CacheType.LRU)
            if details:
                return details.project, UpdateState.UNCHANGED
            # Update from static cache
            details = self.peek(key, CacheType.STATIC)
            if details:
                self._lru_cache[key] = details
                return details.project, UpdateState.UPDATED
        
        # Not found in cache, update the actual entry
        update = self.update_entry(key)
        return self._lru_cache[key].project, update

    def prepare_project(self, project: QgsProject ) -> None:
        """ Set project configuration
        """
        if self._disable_owsurls:
            # Disable ows urls defined in project
            # May be needed because it overrides 
            # any proxy settings
            project.writeEntry("WMSUrl","/", "")
            project.writeEntry("WFSUrl","/", "")
            project.writeEntry("WCSUrl","/", "")
            project.writeEntry("WMTSUrl","/", "")

    def read_project(self, uri: str) -> QgsProject:
        """ Read project from path

            May be used by protocol-handlers to instanciate project
            from uri.
        """
        LOGGER.debug("Reading Qgis project %s", uri)

        # see https://github.com/qgis/QGIS/pull/49266
        if Qgis.QGIS_VERSION_INT < 32601:
            project = self._create_project()
        else:
            project = self._create_project(capabilities=Qgis.ProjectCapabilities())

        readflags = QgsProject.ReadFlags()
        if self._trust_layer_metadata:
            readflags |= QgsProject.FlagTrustLayerMetadata
        if self._disable_getprint:
            readflags |= QgsProject.FlagDontLoadLayouts 
        badlayerh = BadLayerHandler()
        project.setBadLayerHandler(badlayerh)
        if not project.read(uri,  readflags):
            raise RuntimeError(f"Failed to read Qgis project {uri}")

        if self._strict_check and not badlayerh.validateLayers(project):
            raise StrictCheckingError

        self.prepare_project(project)
        return project


class BadLayerHandler(QgsProjectBadLayerHandler):

    def __init__(self):
        super().__init__()
        self.badLayerNames = set()

    def handleBadLayers( self, layers ) -> None:
        """ See https://qgis.org/pyqgis/3.0/core/Project/QgsProjectBadLayerHandler.html
        """
        super().handleBadLayers( layers )

        nameElements = (lyr.firstChildElement("layername") for lyr in layers if lyr)
        self.badLayerNames = set(elem.text() for elem in nameElements if elem)

    def validateLayers( self, project: QgsProject ) -> bool:
        """ Check layers
            
            If layers are excluded do not count them as bad layers
            see https://github.com/qgis/QGIS/pull/33668
        """
        if self.badLayerNames:
            LOGGER.error("Found bad layers: %s", self.badLayerNames)
            restricteds = set(QgsServerProjectUtils.wmsRestrictedLayers(project))
            return self.badLayerNames.issubset(restricteds)
        return True


def get_cacheservice() -> QgsCacheManager:
    return componentmanager.get_service(CACHE_MANAGER_CONTRACTID)


def preload_projects_file( path: Path, cacheservice: QgsCacheManager ) ->  int:
    """ Preload projects from configuration file in static cache
    """
    conf_file = Path(path)
    if not conf_file.exists():
        LOGGER.error("%s file do not exists, ignoring preload config", path)
        return 0
    
    loaded_so_far = 0

    # Read the projects, strip comments 
    with conf_file.open() as fp:
        for p in filter(None,(line.strip('\n ').partition('#')[0] for line in fp.readlines())):
            p = p.strip(' ')
            try:
                cacheservice.update_static_entry(p)
            except StrictCheckingError:
                LOGGER.error("Preload: '%s' as invalid layers - strict mode on" , p)
            except PathNotAllowedError:
                LOGGER.error("Preload: '%s' path not allowed", p)
            except FileNotFoundError:
                LOGGER.error("Preload: '%s' not found", p)
            else:
                loaded_so_far += 1
                LOGGER.info("Preload: '%s' loaded", p)

    return loaded_so_far


def preload_projects() -> None:
    """ Preload projects in cache
    """
    confpath = confservice['projects.cache'].get('preload_config', fallback=None)
    if not confpath:
        return

    preload_projects_file( confpath, get_cacheservice() )


def get_project_summary( key: str, project: QgsProject ):
    """ Return json summary for cached project
    """
    def layer_summary( layer_id: str, layer: QgsMapLayer ):
        return dict(
            id=layer_id,
            name=layer.name(),
            source=layer.publicSource(),
            crs=layer.crs().userFriendlyIdentifier(),
            valid=layer.isValid(),
            spatial=layer.isSpatial(),
        )

    layers = [layer_summary(idstr,l) for (idstr,l) in project.mapLayers().items()]

    return dict(
        cache_key=key,
        filename=project.fileName(),
        bad_layers_count=sum(1 for ls in layers if not ls['valid']),
        layers=layers,
        crs=project.crs().userFriendlyIdentifier(),
        last_modified=project.lastModified().toString(Qt.ISODate)
    )



#
# Copyright 2022 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
""" Store monitoring data in memory table

    Used for testing purposes
    DO NOT USE IN PRODUCTION
"""
import logging

from typing import Dict

from  .base import MonitorBase

LOGGER = logging.getLogger('SRVLOG')

class Monitor(MonitorBase):

    def __init__(self):
        super().__init__()

        self.messages = []

    def emit( self, params: Dict[str,str], meta: Dict ) -> None:
        """ Publish monitor data
        """
        data = dict(self.global_tags)
        data.update(params)
        self.messages.append((data,meta))

_instance = Monitor()

# Entrypoint
def initialize() -> Monitor:
    return _instance




""" AMQP monitor for qgis requests
"""
import asyncio
import logging
import traceback
import json
import os

from typing import Dict, Optional, Any

from amqpclient.concurrent import AsyncPublisher

from ..config  import confservice
from .base  import MonitorBase

LOGGER = logging.getLogger('SRVLOG')

def _read_credentials( vhost: str, user: str ) -> Optional:  # ['PlainCredentials']
    """ Read credentials from passfile
    """
    credential_file = os.getenv("AMQPPASSFILE")
    if not (credential_file and os.path.exists(credential_file)):
        return

    from pika import PlainCredentials

    LOGGER.debug("Using passfile %s", credential_file)
    with open(credential_file) as fp:
        for line in fp.readlines():
            credentials = line.strip()
            if credentials and not credentials.startswith('#'):
                cr_vhost, cr_user, passwd = credentials.split(':')
                if cr_vhost in ('*',vhost) and  cr_user in ('*',user):
                    LOGGER.info("Using credentials for user '%s' on vhost '%s'", user, vhost)
                    return PlainCredentials(user,passwd)


class Monitor(MonitorBase):

    def __init__(self, amqp_client: 'AsyncPublisher', routing_key: str, default_routing: Optional[str]=None ) -> None: # noqa: F821
        """ Init AMQP monitor
        """
        super().__init__()

        self._client = amqp_client

        self._dynamic_routing = routing_key.startswith('@')
        if self._dynamic_routing:
            self._routing_key = routing_key[1:]
            self._default_routing = default_routing
            if not self._default_routing:
                LOGGER.warning("No default routing defined as fallback for dynamic routing")
        else:
            self._routing_key = routing_key

    def emit( self, params: Dict[str,Any], meta: Dict ) -> None:
        """ Publish monitor data
        """
        if self._dynamic_routing:
            try:
                routing_key = self._routing_key.format(META=meta)
            except KeyError: 
                # FALLBACK to default routing key
                if self._default_routing:
                    routing_key = self._default_routing
        else:
            routing_key = self._routing_key

        # Send all params to our logger
        data = dict(self.global_tags, ROUTING_KEY=routing_key)
        data.update(params)
        log_msg = json.dumps(data)
        self._client.publish( log_msg ,
                              routing_key  = routing_key,
                              expiration   = 3000,
                              content_type = 'application/json',
                              content_encoding ='utf-8')

    @classmethod
    def initialize(cls) -> 'Monitor':
        """ Register an instance of Monitor client
        """
        if hasattr(cls,'_instance'):
            return cls._instance

        conf = confservice['monitor:amqp']
        routing_key = conf.get('routing_key')
        if not routing_key:
            return 

        hosts = conf['host']
        user  = conf['user']
        vhost = conf['vhost']
        port  = conf['port']

        reconnect_delay = conf.getint('reconnect_delay')

        kwargs = {}

        if user:
            credentials = _read_credentials( vhost, user )
            if credentials:
                kwargs['credentials'] = credentials

        exchange = conf['exchange']

        client = AsyncPublisher(host=hosts,port=int(port),virtual_host=vhost,
                                reconnect_delay=reconnect_delay,
                                logger=LOGGER, **kwargs)

        # Catch exception in connection
        async def connect():
            try:
                await client.connect(exchange=exchange,exchange_type='topic')
                LOGGER.info("AMQP logger initialized.")
            except Exception:
                LOGGER.error("Failed to initialize AMQP logger: %s",traceback.format_exc())

        asyncio.ensure_future( connect() )

        # Get default routing key in case as fallback in case
        # We fail to get dynamic key
        default_routing_key = conf.get('default_routing_key', fallback=None)

        inst = cls(client, routing_key, default_routing=default_routing_key)
        setattr(cls,'_instance', inst)
        return inst




#
# Copyright 2018 3liz
# Author: David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import logging

from .basehandler import BaseHandler
from ..version import __version__
from ..config import confservice

LOGGER=logging.getLogger('SRVLOG')


class LandingPage(BaseHandler):

    def initialize(self):
        super().initialize()

        self._metadata = confservice['metadata']

    def get(self):
        """
        """
        req = self.request
        self._server = f"{req.protocol}://{req.host}"
        doc = {}
        doc.update(
            servers=[{
                'url': self._server,
            }],
            info=self.service_infos(),
            externalDocs=self.external_doc(),
            paths={},
        )
        self.write_json(doc)

    def service_infos(self):
        _m = self._metadata.get
        doc = {
            'title': _m('title'),
            'description': _m('description'),
            'termsOfService': _m('terms_of_service'),
            'contact': {
                'name': _m('contact_name'),
                'url': _m('contact_url'),
                'email': _m('contact_email'),
            },
            'licence': {
                'name': _m('licence_name'),
                'url': _m('licence_url'),
            },
            'version': __version__,
        }

        doc.update(self.qgis_version_info())
        return doc

    def qgis_version_info(self):
        try:
            from qgis.core import Qgis
            qgis_version = Qgis.QGIS_VERSION_INT
            qgis_release = Qgis.QGIS_RELEASE_NAME  
        except ImportError:
            LOGGER.critical("Failed to import Qgis module !")
            qgis_version = qgis_release = 'n/a'

        return (
            ('x-qgis-version', qgis_version),
            ('x-qgis-release', qgis_release),   
        )  

    def external_doc(self):
        return {
            'description': self._metadata['external_doc_description'],
            'url': self._metadata['external_doc_url'],
        }




#
# Copyright 2021 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
    Wrapper around more tornado-like request handler for
    implementing Qgis server API
"""
import logging

from tornado.web import HTTPError # noqa F401

from pyqgisserver.qgscache.cachemanager import (get_cacheservice, get_project_summary)
 
from .handler import RequestHandler, register_handlers

LOGGER = logging.getLogger('SRVLOG')


class CacheCollection(RequestHandler):

    def get(self, key: str=None) -> None: 
        """ Return project cache info
        """
        if not key:
            # Try to get key from param
            key = self.request.parameter('MAP')

        if not key:
            raise HTTPError(400,reason="Missing project specification")

        cache = get_cacheservice()
        project, _ = cache.lookup(key, refresh=False)
 
        self.write(get_project_summary(key, project))  
    
 
def register( serverIface ):
    """ Register plugins api handlers
    """
    register_handlers(serverIface, "/cache","CacheManagment",
                      [
                          (r'/content/(?P<key>.+)$', CacheCollection),
                          (r'/', CacheCollection),
                      ])




#
# Copyright 2018 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import logging
import traceback

import tornado.web

from tornado.web import HTTPError
from tornado.routing import Router

from typing import TypeVar, List, NamedTuple

from .handlers import ErrorHandler

from pyqgisservercontrib.core.filters import _FilterBase

HandlerDelegate = TypeVar('HandlerDelegate')

LOGGER = logging.getLogger('SRVLOG')

class _Policy(NamedTuple):
    pri: int
    filters: List[_FilterBase]


def load_access_policies() -> List[_FilterBase]:
    """ Create filter list
    """
    collection = []

    class policy_service:
        @staticmethod
        def add_filters(filters, pri=0):
            if LOGGER.isEnabledFor(logging.DEBUG):
                LOGGER.debug("Adding policy filter(s):\n%s", '\n'.join(str(f) for f in filters)) 
            collection.append(_Policy(pri, [p for p in filters if isinstance(p,_FilterBase)]))

    import pyqgisservercontrib.core.componentmanager as cm
    cm.register_entrypoints('py_qgis_server.access_policy', policy_service) 

    # Sort filters
    collection.sort(key=lambda p: p.pri, reverse=True)
    return collection


class MiddleWareRouter(Router):

    def __init__(self, app: tornado.web.Application) -> None:
        """ Initialize middleware filters
        """
        LOGGER.info("Initializing middleware filters")
        self.app = app
        self.policies = load_access_policies()

    def find_handler(self, request, **kwargs) -> HandlerDelegate:
        """ Define middleware prerocessing 
        """
        # Find matching paths
        for policy in self.policies:
            for filt in policy.filters:
                # Find the first filter that match the path
                match, path = filt.match(request.path)
                if match:
                    LOGGER.debug("Found matching filter for %s -> %s:\n%s", request.path, path, filt)
                    request.path = path
                    try:
                        filt.apply(request)
                        break
                    except HTTPError as err:
                        kwargs = {'status_code': err.status_code, 'reason': err.reason}
                        return self.app.get_handler_delegate(request, ErrorHandler, kwargs)
                    except Exception:
                        LOGGER.critical(traceback.format_exc())
                        return self.app.get_handler_delegate(request, ErrorHandler, {'status_code': 500})

        return self.app.find_handler(request, **kwargs)




#
# Copyright 2018-2019 3liz
# Author: David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" Qgis server handler
"""
import logging

from .asynchandler import AsyncClientHandler

from typing import Awaitable

LOGGER = logging.getLogger('SRVLOG')


class OAPIHandler(AsyncClientHandler):
    """ Handle Qgis api
    """
    def initialize(self, service: str, **kwargs) -> None:
        super().initialize(**kwargs )
        self.ogc_scheme = 'OAF'
        self._service_name = service.upper()

    def prepare(self) -> None:
        super().prepare()
        # Replace MAP key with uppercase
        args = self.request.arguments
        if 'MAP' in args:
            return 
        for k in args:
            if k.upper() == 'MAP':
                key = k
                val = args[k]
                break
        else:
            return
        del args[key]
        args['MAP'] = val

    async def delete(self) -> Awaitable[None]:
        await self.handle_request('DELETE')

    async def put(self) -> Awaitable[None]:
        await self.handle_request('PUT')

    async def patch(self) -> Awaitable[None]:
        await self.handle_request('PATCH')

    async def options(self) -> Awaitable[None]:
        await self.handle_request('OPTIONS')

    def get_monitor_params( self ) -> None:
        """ Override
        """
        params = dict(
            MAP = self.request.arguments.get('MAP','__unknown__'),
            SERVICE = self._service_name,
            REQUEST = self.request.path,
        )
        return params






from .basehandler import ( # noqa F401
    BaseHandler, 
    NotFoundHandler,
    ErrorHandler
)
from .statushandler import ( # noqa F401 
    StatusHandler, 
    PingHandler
)
from .owshandler import ( # noqa F401
    OwsHandler, 
)

from .oapihandler import ( # noqa F401
    OAPIHandler
)

from .landingpage import ( # noqa F401
    LandingPage
)

